@@ -1,13 +1,13 @@
 import itertools
 import typing
-from collections import defaultdict
+import uuid
+from collections import deque
 
 import attr
 from transitions import Machine
 
-from applitools.common import Region, RenderStatus, logger
+from applitools.common import RenderStatus, logger
 
-from .render_task import RenderTask
 from .vg_task import VGTask
 
 if typing.TYPE_CHECKING:
@@ -23,43 +23,155 @@ if typing.TYPE_CHECKING:
     from applitools.selenium.fluent import SeleniumCheckSettings
 
     from .eyes_connector import EyesConnector
-    from .visual_grid_runner import VisualGridRunner
 
 NEW = "new"
-NOT_RENDERED = "not_rendered"
-RENDERED = "rendered"
+NOT_OPENED = "not_opened"
 OPENED = "opened"
 COMPLETED = "completed"
 TESTED = "tested"
 
-STATES = [NEW, NOT_RENDERED, OPENED, RENDERED, COMPLETED, TESTED]
+STATES = [NEW, OPENED, NOT_OPENED, COMPLETED, TESTED]
 TRANSITIONS = [
-    {"trigger": "becomes_not_rendered", "source": NEW, "dest": NOT_RENDERED},
-    {"trigger": "becomes_opened", "source": RENDERED, "dest": OPENED},
-    {"trigger": "becomes_rendered", "source": NOT_RENDERED, "dest": RENDERED},
+    {"trigger": "becomes_not_opened", "source": NEW, "dest": NOT_OPENED},
+    {"trigger": "becomes_opened", "source": NOT_OPENED, "dest": OPENED},
     {
         "trigger": "becomes_tested",
-        "source": [NEW, NOT_RENDERED, OPENED],
+        "source": [NEW, NOT_OPENED, OPENED],
         "dest": TESTED,
     },
     {
         "trigger": "becomes_completed",
-        "source": [NEW, NOT_RENDERED, RENDERED, OPENED, TESTED],
+        "source": [NEW, NOT_OPENED, OPENED, TESTED],
         "dest": COMPLETED,
     },
 ]
 
 
+@attr.s(hash=False, str=False)
+class RunningTestCheck(object):
+    name = attr.ib()  # type: Text
+    uuid = attr.ib(init=False, repr=False, factory=lambda: str(uuid.uuid4()))
+    running_test = attr.ib(repr=False)  # type: RunningTest
+
+    check_settings = attr.ib()  # type: SeleniumCheckSettings
+    region_selectors = attr.ib()  # type: List[List[VisualGridSelector]]
+    source = attr.ib()  # type: Text
+
+    regions = attr.ib(init=False, factory=list)
+    queue = attr.ib(init=False, factory=deque)
+
+    def set_render_request(self, render_request):
+        short_description = "{} of {}".format(
+            self.running_test.configuration.test_name,
+            self.running_test.configuration.app_name,
+        )
+        tag = self.check_settings.values.name
+        render_task = self._render_task(tag, short_description, render_request)
+        self.queue = deque([render_task])
+
+    def __hash__(self):
+        return hash(self.name + self.uuid)
+
+    def _render_task(self, tag, short_description, render_request):
+        def render_task_succeeded(render_status):
+            # type: (List[RenderStatusResults]) -> None
+            logger.debug(
+                "render_task_succeeded: task.uuid: {}".format(render_task.uuid)
+            )
+            if render_status:
+                self.running_test.eyes.render_status_for_task(
+                    render_task.uuid, render_status
+                )
+                if render_status.status == RenderStatus.RENDERED:
+                    for vgr in render_status.selector_regions:
+                        if vgr.error:
+                            logger.error(vgr.error)
+                        else:
+                            self.regions.append(vgr.to_region())
+                    logger.debug(
+                        "render_task_succeeded: uuid: {}\n\tregions {}".format(
+                            render_task.uuid, self.regions
+                        )
+                    )
+                    # schedule check task
+                    self.queue.append(self._check_task(render_task, tag))
+                elif render_status and render_status.status == RenderStatus.ERROR:
+                    self.running_test.task_queue.clear()
+                    self.running_test.open_queue.clear()
+                    self.running_test.close_queue.clear()
+                    self.running_test.watch_open = {}
+                    self.running_test.watch_task = {}
+                    self.running_test.watch_close = {}
+
+                    self.running_test.abort()
+                    self.running_test.becomes_tested()
+            else:
+                logger.error(
+                    "Wrong render status! Render returned status {}".format(
+                        render_status
+                    )
+                )
+                self.running_test.becomes_completed()
+
+        def render_task_error(e):
+            logger.debug(
+                "render_task_error: task.uuid: {}\n{}".format(render_task.uuid, str(e))
+            )
+            self.running_test.pending_exceptions.append(e)
+            self.running_test.becomes_completed()
+
+        render_task = VGTask(
+            "RunningTest.render {} - {}".format(short_description, tag),
+            lambda: self.running_test.rendering_service.render(
+                render_request, render_task_succeeded, render_task_error
+            ),
+        )
+        return render_task
+
+    def _check_task(self, render_task, tag):
+        def check_run():
+            logger.debug("check_run: render_task.uuid: {}".format(render_task.uuid))
+            self.running_test.eyes.check(
+                self.check_settings,
+                render_task.uuid,
+                self.region_selectors,
+                self.regions,
+                self.source,
+            )
+
+        check_task = VGTask(
+            "perform check {} {}".format(tag, self.check_settings), check_run
+        )
+
+        def check_task_completed():
+            # type: () -> None
+            logger.debug("check_task_completed: task.uuid: {}".format(check_task.uuid))
+            if (
+                self.running_test.task_lock
+                and self.running_test.task_lock.uuid == self.uuid
+            ):
+                self.running_test.task_lock = None
+
+            self.running_test.watch_task[self] = True
+            if self.running_test.all_tasks_completed(self.running_test.watch_task):
+                self.running_test.becomes_tested()
+
+        def check_task_error(e):
+            logger.debug("check_task_error: task.uuid: {}".format(check_task.uuid))
+            self.running_test.pending_exceptions.append(e)
+            self.running_test.becomes_completed()
+
+        check_task.on_task_completed(check_task_completed)
+        check_task.on_task_error(check_task_error)
+        return check_task
+
+
 @attr.s(hash=True, str=False)
 class RunningTest(object):
     eyes = attr.ib(hash=False, repr=False)  # type: EyesConnector
     configuration = attr.ib(hash=False, repr=False)  # type: Configuration
     browser_info = attr.ib()  # type: RenderBrowserInfo
-    # listener = attr.ib()  # type:
-    region_selectors = attr.ib(
-        init=False, factory=list, hash=False
-    )  # type: List[List[VisualGridSelector]]
-    regions = attr.ib(init=False, factory=lambda: defaultdict(list), hash=False)
+    rendering_service = attr.ib()
 
     tasks_list = attr.ib(init=False, factory=list, hash=False)
     test_uuid = attr.ib(init=False)
@@ -79,17 +191,15 @@ class RunningTest(object):
 
     def _initialize_vars(self):
         # type: () -> None
-        self.open_queue = []  # type: List[VGTask]
-        self.task_queue = []  # type: List[VGTask]
-        self.render_queue = []  # type: List[RenderTask]
-        self.close_queue = []  # type: List[VGTask]
+        self.open_queue = deque()  # type: deque[VGTask]
+        self.task_queue = deque()  # type: deque[RunningTestCheck]
+        self.close_queue = deque()  # type: deque[VGTask]
         self.watch_open = {}  # type: Dict[VGTask, bool]
-        self.watch_task = {}  # type: Dict[VGTask, bool]
-        self.watch_render = {}  # type: Dict[RenderTask, bool]
+        self.watch_task = {}  # type: Dict[RunningTestCheck, bool]
         self.watch_close = {}  # type: Dict[VGTask, bool]
         self.task_lock = None  # type: Optional[VGTask]
         self.test_result = None  # type: Optional[TestResults]
-        self.pending_exceptions = []  # type: List[Exception]
+        self.pending_exceptions = deque()  # type: deque[Exception]
 
     def _initialize_state_machine(self):
         # type: () -> None
@@ -108,41 +218,25 @@ class RunningTest(object):
 
     @property
     def queue(self):
-        # type: () -> List
+        # type: () -> deque[VGTask]
         if self.state == NEW:
-            return []
-        elif self.state == NOT_RENDERED:
-            return self.render_queue
-        elif self.state == RENDERED:
+            return deque()
+        elif self.state == NOT_OPENED:
             return self.open_queue
         elif self.state == OPENED:
             if self.task_lock:
-                return []
+                return self.task_lock.queue
             elif self.task_queue:
-                self.task_lock = self.task_queue[-1]
-                return self.task_queue
+                self.task_lock = self.task_queue.popleft()
+                return self.task_lock.queue
+            elif self.close_queue:
+                # in case no checks, but close scheduled
+                return self.close_queue
+            return deque()
         elif self.state == TESTED:
             return self.close_queue
         elif self.state == COMPLETED:
-            return []
-        else:
-            raise TypeError("Unsupported state")
-
-    @property
-    def score(self):
-        # type: () -> int
-        if self.state == NEW:
-            return 0
-        elif self.state == NOT_RENDERED:
-            return len(self.render_queue) * 10
-        elif self.state == RENDERED:
-            return len(self.open_queue)
-        elif self.state == OPENED:
-            return len(self.task_queue)
-        elif self.state == TESTED:
-            return len(self.close_queue)
-        elif self.state == COMPLETED:
-            return 0
+            return deque()
         else:
             raise TypeError("Unsupported state")
 
@@ -159,6 +253,9 @@ class RunningTest(object):
             logger.debug("open_task_succeeded: task.uuid: {}".format(open_task.uuid))
             self.watch_open[open_task] = True
             if self.all_tasks_completed(self.watch_open):
+                if self.state == TESTED:
+                    logger.debug("open_task_succeeded: test session was aborted")
+                    return
                 self.becomes_opened()
 
         def open_task_error(e):
@@ -175,159 +272,33 @@ class RunningTest(object):
 
     def check(
         self,
-        tag,  # type: Text
         check_settings,  # type: SeleniumCheckSettings
-        script_result,  # type: Dict[str, Any]
-        visual_grid_manager,  # type: VisualGridRunner
-        region_selectors,
-        region_to_check,
-        script_hooks,
+        region_selectors,  # type: List[List[VisualGridSelector]]
         source,  # type: Optional[Text]
     ):
-        # type: (...) -> None
-        logger.debug("RunningTest %s , %s" % (tag, check_settings))
-        render_task = self._render_task(
-            script_result,
-            tag,
-            visual_grid_manager,
-            region_selectors,
-            region_to_check,
-            script_hooks,
+        # type: (...) -> RunningTestCheck
+        logger.debug("RunningTest %s" % check_settings)
+        running_test_check_task = RunningTestCheck(
+            "RunningTestCheck({})".format(check_settings.values.name),
+            self,
             check_settings,
+            region_selectors,
+            source,
         )
 
-        def check_run():
-            logger.debug("check_run: render_task.uuid: {}".format(render_task.uuid))
-            self.eyes.check(
-                tag,
-                check_settings,
-                render_task.uuid,
-                region_selectors,
-                self.regions[render_task],
-                source,
-            )
-
-        check_task = VGTask(
-            "perform check {} {}".format(tag, check_settings), check_run
-        )
-
-        def check_task_completed():
-            # type: () -> None
-            logger.debug("check_task_completed: task.uuid: {}".format(check_task.uuid))
-            self.watch_task[check_task] = True
-            if self.task_lock and self.task_lock.uuid == check_task.uuid:
-                self.task_lock = None
-            if self.all_tasks_completed(self.watch_task):
-                self.becomes_tested()
-
-        def check_task_error(e):
-            logger.debug("check_task_error: task.uuid: {}".format(check_task.uuid))
-            self.pending_exceptions.append(e)
-
-        check_task.on_task_completed(check_task_completed)
-        check_task.on_task_error(check_task_error)
-        self.task_queue.insert(0, check_task)
-        self.watch_task[check_task] = False
-
-    def _render_task(
-        self,
-        script_result,  # type: Dict[Text, Any]
-        tag,  # type: Text
-        visual_grid_manager,  # type: VisualGridRunner
-        region_selectors,  # type: List
-        region_to_check,  # type: Region
-        script_hooks,  # type: Dict[Text, Any]
-        check_settings,
-    ):
-        # type: (...) -> RenderTask
-        short_description = "{} of {}".format(
-            self.configuration.test_name, self.configuration.app_name
-        )
-        render_task = RenderTask(
-            name="RunningTest.render {} - {}".format(short_description, tag),
-            script=script_result,
-            resource_cache=visual_grid_manager.resource_cache,
-            put_cache=visual_grid_manager.put_cache,
-            rendering_info=self.eyes.render_info(),
-            eyes_connector=self.eyes,
-            region_selectors=region_selectors,
-            size_mode=check_settings.values.size_mode,
-            region_to_check=region_to_check,
-            script_hooks=script_hooks,
-            agent_id=self.eyes.base_agent_id,
-            selector=check_settings.values.selector,
-            request_options=self._options_dict(
-                self.configuration.visual_grid_options,
-                check_settings.values.visual_grid_options,
-            ),
-        )
-        logger.debug("RunningTest %s" % render_task.name)
-        render_index = render_task.add_running_test(self)
-
-        def render_task_succeeded(render_statuses):
-            # type: (List[RenderStatusResults]) -> None
-            logger.debug(
-                "render_task_succeeded: task.uuid: {}".format(render_task.uuid)
-            )
-            render_status = render_statuses[render_index]
-            if render_status:
-                if not render_status.device_size:
-                    render_status.device_size = self.browser_info.viewport_size
-                self.eyes.render_status_for_task(render_task.uuid, render_status)
-                if render_status.status == RenderStatus.RENDERED:
-                    for vgr in render_status.selector_regions:
-                        if vgr.error:
-                            logger.error(vgr.error)
-                        else:
-                            self.regions[render_task].append(vgr.to_region())
-                    self.watch_render[render_task] = True
-                    if self.all_tasks_completed(self.watch_render):
-                        self.becomes_rendered()
-                    logger.debug(
-                        "render_task_succeeded: uuid: {}\n\tregions {}".format(
-                            render_task.uuid, self.regions.get(render_task)
-                        )
-                    )
-                elif render_status and render_status.status == RenderStatus.ERROR:
-                    self.watch_render[render_task] = True
-                    del self.task_queue[:]
-                    del self.open_queue[:]
-                    del self.close_queue[:]
-                    self.watch_open = {}
-                    self.watch_task = {}
-                    self.watch_close = {}
-                    self.abort()
-                    if self.all_tasks_completed(self.watch_render):
-                        self.becomes_tested()
-            else:
-                logger.error(
-                    "Wrong render status! Render returned status {}".format(
-                        render_status
-                    )
-                )
-                self.becomes_completed()
-
-        def render_task_error(e):
-            logger.debug(
-                "render_task_error: task.uuid: {}\n{}".format(render_task.uuid, str(e))
-            )
-            self.pending_exceptions.append(e)
-            self.becomes_completed()
-
-        render_task.on_task_succeeded(render_task_succeeded)
-        render_task.on_task_error(render_task_error)
-        self.render_queue.append(render_task)
-        self.watch_render[render_task] = False
-        return render_task
+        self.task_queue.append(running_test_check_task)
+        self.watch_task[running_test_check_task] = False
+        return running_test_check_task
 
     def close(self):
         # type: () -> Optional[Any]
         if self.state == NEW:
             self.becomes_completed()
             return None
-        elif self.state in [NOT_RENDERED, RENDERED, OPENED, TESTED]:
+        elif self.state in [NOT_OPENED, OPENED, TESTED]:
             close_task = VGTask(
-                "close {}".format(self.browser_info), lambda: self.eyes.close(False)
+                "close {}".format(self.browser_info),
+                lambda: self.eyes.is_open and self.eyes.close(False),
             )
             logger.debug("RunningTest %s" % close_task.name)
 
@@ -335,9 +306,11 @@ class RunningTest(object):
                 logger.debug(
                     "close_task_succeeded: task.uuid: {}".format(close_task.uuid)
                 )
-                self.test_result = test_result
-                if callable(self.on_results):
-                    self.on_results(test=self, test_result=test_result)
+                # abort() could add test_result already
+                if self.test_result is None and test_result:
+                    self.test_result = test_result
+                    if callable(self.on_results):
+                        self.on_results(test=self, test_result=test_result)
 
             def close_task_completed():
                 # type: () -> None
@@ -368,7 +341,9 @@ class RunningTest(object):
             return None
 
         def ensure_and_abort():
-            self.eyes._ensure_running_session()
+            if not self.eyes.is_open:
+                # open new session if no opened
+                self.eyes._ensure_running_session()
             return self.eyes.abort()
 
         abort_task = VGTask("abort {}".format(self.browser_info), ensure_and_abort)
